@@ -22,7 +22,12 @@ from os import path
 import pythomics.templates as templates
 from pythomics.proteomics.structures import PeptideObject, ScanObject, Chromatogram
 from pythomics.proteomics import config
-import re, os, sqlite3, StringIO, zipfile, time
+import re, os, sqlite3, zipfile, time
+import six
+
+if six.PY3:
+    xrange = range
+from six import string_types
 try:
     from lxml import etree
 except ImportError:
@@ -38,10 +43,30 @@ scanSplitter = re.compile(r'[\t\s]')
 distillerParse = re.compile(r'_DISTILLER_RAWFILE\[(\d+)\]=\(1\)(.+)')
 lastSplit = re.compile(r'.+[/\\](.+)')
 
+# def line_profiler(view=None, extra_view=None):
+#     import line_profiler
+#
+#     def wrapper(view):
+#         def wrapped(*args, **kwargs):
+#             prof = line_profiler.LineProfiler()
+#             prof.add_function(view)
+#             if extra_view:
+#                 [prof.add_function(v) for v in extra_view]
+#             with prof:
+#                 resp = view(*args, **kwargs)
+#             prof.print_stats()
+#             return resp
+#         return wrapped
+#     if view:
+#         return wrapper(view)
+#     return wrapper
+
 class GenericProteomicIterator(object):
     """
     Generics to be overridden
     """
+    chromatogram = None
+
     def getChromatogram(self):
         pass
 
@@ -94,17 +119,17 @@ class GuessIterator(templates.GenericIterator):
     def __init__(self, *args, **kwargs):
         super(GuessIterator, self).__init__(*args, **kwargs)
         filename = self.filename.name.lower()
-        if '.mzml' in filename:
+        if filename.endswith('.mzml'):
             self.parser = MZMLIterator
-        elif '.pepxml' in filename or '.pep.xml' in filename:
+        elif filename.endswith('.pepxml') or filename.endswith('.pep.xml'):
             self.parser = PepXMLIterator
-        elif '.xml' in filename:
+        elif filename.endswith('.xml'):
             self.parser = XTandemXMLIterator
-        elif '.mgf' in filename:
+        elif filename.endswith('.mgf'):
             self.parser = MGFIterator
-        elif '.dat' in filename:
+        elif filename.endswith('.dat'):
             self.parser = MascotDATIterator
-        elif '.msf' in filename or '.pdresult' in filename:
+        elif filename.endswith('.msf') or filename.endswith('.pdresult'):
             self.parser = ThermoMSFIterator
         self.parser = self.parser(*args, **kwargs)
         private = {'__iter__',}
@@ -131,11 +156,17 @@ class MZMLIterator(XMLFileNameMixin, templates.GenericIterator, GenericProteomic
         self.full = full
         self.ms_filter = ms_filter
         self.start = start
-        spectra_tags = ('{http://psi.hupo.org/ms/mzml}spectrum', '{http://psi.hupo.org/ms/mzml}indexedmzML', '{http://psi.hupo.org/ms/mzml}chromatogram')
+        spectra_tags = ('{http://psi.hupo.org/ms/mzml}spectrum', '{http://psi.hupo.org/ms/mzml}chromatogram')
         self.filename.seek(0)
-        self.spectra = etree.parse(self.filename.name).iter(tag=spectra_tags) if self.gzip else etree.iterparse(self.filename.name, tag=spectra_tags)
+        self.spectra = etree.parse(self.filename).iter(tag=('{http://psi.hupo.org/ms/mzml}indexedmzML',)) if self.gzip else etree.iterparse(self.filename, tag=('{http://psi.hupo.org/ms/mzml}indexedmzML',), recover=True)
+        index = next(self.spectra)
+        if index:
+            print('mzML index is missing. Please provide files with an index.')
+            self.ra = dict([(self._get_scan_from_string(i.values()[0]), i.text) for i in index[1].findall('{0}indexList/{0}index/'.format('{http://psi.hupo.org/ms/mzml}'))])
+        else:
+            self.ra = {}
+        self.spectra = etree.parse(self.filename).iter(tag=spectra_tags) if self.gzip else etree.iterparse(self.filename, tag=spectra_tags, recover=True)
         self.scans = {}
-        self.ra = {}
         self.startIter = True
         self.db = None
             
@@ -149,27 +180,23 @@ class MZMLIterator(XMLFileNameMixin, templates.GenericIterator, GenericProteomic
         else:
             array = base64.b64decode(array)
         if '64-bit float' in params:
-            array = [struct.unpack('d', array[i:i+8])[0] for i in xrange(0, len(array), 8)]
+            array = struct.unpack('{}d'.format(len(array)/8), array)
         else:
-            array = [struct.unpack('f', array[i:i+4])[0] for i in xrange(0, len(array), 4)]
+            array = struct.unpack('{}f'.format(len(array)/4), array)
         return array
 
     def parselxml(self, spectra, full=False, namespace='{http://psi.hupo.org/ms/mzml}'):
         try:
-            if spectra.tag == '{0}indexedmzML'.format(namespace):
-                # read our index in
-                self.ra = dict([(self._get_scan_from_string(i.values()[0]), i.text) for i in spectra.findall('{0}indexList/{0}index/'.format(namespace))])
-                return None
-            elif spectra.tag == '{0}spectrum'.format(namespace):
+            if spectra.tag == '{0}spectrum'.format(namespace):
                 scanObj = ScanObject()
                 # spectra_info = dict(zip(spectra.keys(),spectra.values()))
                 spectra_params = dict([(i.get('name'), i.get('value')) for i in spectra.findall('{0}cvParam'.format(namespace))])
                 scan_info = dict([(i.get('name'), i.get('value')) for i in spectra.findall('{0}scanList/{0}scan/{0}cvParam'.format(namespace))])
                 precursor_info = dict([(i.get('name'), i.get('value')) for i in spectra.findall('{0}precursorList/{0}precursor/{0}selectedIonList/{0}selectedIon/{0}cvParam'.format(namespace))])
-                # print 'spectra info', spectra_info
-                # print 'spectra params', spectra_params
-                # print 'our scan info', scan_info
-                # print 'our precursor info', precursor_info
+                parent_scan = spectra.findall('{0}precursorList/{0}precursor'.format(namespace))
+                if parent_scan:
+                    parent_scan = self._get_scan_from_string(parent_scan[0].get('spectrumRef'))
+                    scanObj.parent = parent_scan
                 ms_level = int(spectra_params.get('ms level', 0))
                 charge = precursor_info.get('charge state', 0)
                 precursor_ion = precursor_info.get('selected ion m/z', 0)
@@ -189,7 +216,10 @@ class MZMLIterator(XMLFileNameMixin, templates.GenericIterator, GenericProteomic
                     mzmls = self.unpack_array(mzmls.find('{0}binary'.format(namespace)).text, mzml_params, namespace=namespace)
                     intensity_params = dict([(i.get('name'), i.get('value')) for i in intensities.findall('{0}cvParam'.format(namespace))])
                     intensities = self.unpack_array(intensities.find('{0}binary'.format(namespace)).text, intensity_params, namespace=namespace)
-                    scanObj.scans = zip(mzmls, intensities)
+                    scanObj.scans = list(zip(mzmls, intensities))
+                if self.filetype == 'thermo':
+                    target_info = dict([(i.get('name'), i.get('value')) for i in spectra.findall('{0}precursorList/{0}precursor/{0}isolationWindow/{0}cvParam'.format(namespace))])
+                    scanObj.product_ion = target_info.get('isolation window target m/z', 0)
                 spectra.clear()
                 return scanObj
             elif spectra.tag == '{0}chromatogram'.format(namespace):
@@ -239,7 +269,7 @@ class MZMLIterator(XMLFileNameMixin, templates.GenericIterator, GenericProteomic
                         mzmls = self.unpack_array(mzmls.find('{0}binary'.format(namespace)).text, mzml_params, namespace=namespace)
                         intensity_params = dict([(i.get('name'), i.get('value')) for i in intensities.findall('{0}cvParam'.format(namespace))])
                         intensities = self.unpack_array(intensities.find('{0}binary'.format(namespace)).text, intensity_params, namespace=namespace)
-                        scanObj.scans = [(i,j) for i,j in zip(mzmls, intensities)]
+                        scanObj.scans = list(zip(mzmls, intensities))
                     spectra.clear()
                     return scanObj
         except:
@@ -251,21 +281,16 @@ class MZMLIterator(XMLFileNameMixin, templates.GenericIterator, GenericProteomic
 
     def next(self):
         if self.spectra:
-            spectra = None
-            while spectra is None:
-                try:
-                    spectra = self.spectra.next()
-                except:
-                    pass
+            spectra = next(self.spectra)
         else:
             raise StopIteration
         if self.gzip:
             scan = self.parselxml(spectra, full=True)
             if self.store and isinstance(scan, ScanObject):
                 self.scans[scan.id] = scan
-            return scan
         else:
-            return self.parselxml(spectra[1], full=self.full)
+            scan = self.parselxml(spectra[1], full=self.full)
+        return scan
 
     def getChromatogram(self):
         return self.chromatogram
@@ -283,14 +308,20 @@ class MZMLIterator(XMLFileNameMixin, templates.GenericIterator, GenericProteomic
                 if id in self.scans:
                     return self.scans[id]
             else:
-                if not self.ra:
-                    [i for i in self]
+                while id not in self.ra:
+                    spectra = self.next()
+                    if spectra.id == id:
+                        return spectra
                 self.filename.seek(int(self.ra[str(id)]))
                 entry = self.filename.readline()
-                opening_tag = entry[1:entry.find(' ')]
-                closing_tag = '</{}'.format(opening_tag)
+                opening_tag = '<spectrum '
+                closing_tag = '</spectrum>'
+                while entry and opening_tag not in entry:
+                    entry = self.filename.readline()
+                if not entry:
+                    return None
                 row = [entry]
-                while closing_tag not in entry:
+                while entry and closing_tag not in entry:
                     entry = self.filename.readline()
                     row.append(entry)
                 spectra = etree.fromstring(''.join(row))
@@ -377,7 +408,6 @@ class PepXMLIterator(XMLFileNameMixin, GenericProteomicIterator, templates.Gener
             for i in analysis:
                 info = dict(i.items())
                 analysis_type = info.get('analysis', False)
-                # print analysis_type
                 if analysis_type == 'peptideprophet':
                     results = i.find('{0}peptideprophet_result'.format(namespace))
                     prophet_info = dict(results.items())
@@ -390,7 +420,6 @@ class PepXMLIterator(XMLFileNameMixin, GenericProteomicIterator, templates.Gener
                     results = i.find('{0}ptmprophet_result'.format(namespace))
                     ptm_info = dict(results.items())
                     probs = results.findall('{0}mod_aminoacid_probability'.format(namespace))
-                    # print ptm_info
                     mod_mass = self.ptm_regex.match(ptm_info['ptm']).group(1)
                     for prob in probs:
                         prob_info = dict(prob.items())
@@ -419,8 +448,6 @@ class PepXMLIterator(XMLFileNameMixin, GenericProteomicIterator, templates.Gener
                 mods = search_result.findall('{0}mod_aminoacid_mass'.format(namespace))
                 for i in mods:
                     mod_info = dict(i.items())
-                    # print i.items()
-                    # print mod_info
                     mod_pos = int(mod_info['position'])-1
                     mod_aa = peptide[mod_pos]
                     pepObj.addModification(mod_aa, mod_pos, mod_info['mass'], mod_info['mass'])
@@ -781,7 +808,6 @@ class MGFIterator(templates.GenericIterator, GenericProteomicIterator):
         foundTitle = False
         scanObj = ScanObject()
         scanObj.ms_level = 2
-#        print 'stuff in scan',scan
         for row in scan.split('\n'):
             if not row:
                 continue
@@ -971,7 +997,6 @@ class MascotDATIterator(templates.GenericIterator, GenericProteomicIterator):
         modind=1
         mod = self.resfile.getSectionValueStr(msparser.ms_mascotresfile.SEC_MASSES, "delta%d"%modind)
         while mod:
-            # print mod
             modind+=1
             mod = self.resfile.getSectionValueStr(msparser.ms_mascotresfile.SEC_MASSES, "delta%d"%modind)
         self.aahelper = msparser.ms_aahelper(self.resfile, "enzymes.txt")
@@ -1077,8 +1102,6 @@ class MascotDATIterator(templates.GenericIterator, GenericProteomicIterator):
                         chargeRule = 0
                         matched = {'labels':[],'m/z': [], 'intensity': [], 'error': [], 'series': [], 'start': [], 'end': [], 'losses': [], 'charge': []}
                         for rule in (int(arule) for arule in rules):
-#                            print 'looking at rule',rule
-#                            print 'crules',self.cRules
                             if rule == 1:
                                 chargeRule = 1
                             elif rule == 2 or rule == 3:
@@ -1086,11 +1109,9 @@ class MascotDATIterator(templates.GenericIterator, GenericProteomicIterator):
                                     chargeRule = 2
                             else:
                                 for chargeState in xrange(charge+1):
-#                                    print 'checking chargestate',chargeState,'with chargerule',chargeRule,'on rule',rule
                                     if (chargeState == 1 and chargeRule == 1) or (chargeState == 2 and chargeRule == 2) and rule not in self.cRules:
                                         frags = msparser.ms_fragmentvector()
                                         errs = msparser.ms_errs()
-                                        #print pep, rule, chargeState, 0, pep.getMrCalc(), self.sparams.getMassType(), frags, errs
                                         self.aahelper.setMasses(self.massFile)
                                         self.aahelper.setAvailableModifications(self.vecFixed,self.vMods)
                                         fres = self.aahelper.calcFragmentsEx(pep, rule, chargeState, 0, pep.getMrCalc(), self.sparams.getMassType(), frags, errs)
@@ -1126,7 +1147,6 @@ class MascotDATIterator(templates.GenericIterator, GenericProteomicIterator):
                                                 fragname = ""
                                             sys.stderr.write('fragment name%s\n'%fragname)
                         scanObj.matched = matched
-#                        print matched
                     specId = self.specParse.search(stitle).group(1)
                     scanObj.id = specId
                     self.scanMap[(specId, peptide)] = scanObj
@@ -1143,7 +1163,7 @@ class ThermoMSFIterator(templates.GenericIterator, GenericProteomicIterator):
         self.clvl = clvl
         self.full = full
         self.srank = srank
-        if isinstance(filename,(str,unicode)):
+        if isinstance(filename, string_types):
             self.f = open(filename, 'rb')
         else:
             raise Exception(TypeError,"Unknown Type of filename -- must be a file path")
@@ -1180,7 +1200,7 @@ class ThermoMSFIterator(templates.GenericIterator, GenericProteomicIterator):
         self.scans = []
         self.scanTable = {}
         for i in self.cur.fetchall():
-            self.modTable[i[0]] = (i[1],i[2])
+            self.modTable[str(i[0])] = (i[1],i[2])
         self.aaTable = dict([(i[0],i[1]) for i in self.conn.execute("select aa.AminoAcidID, aa.OneLetterCode from aminoacids aa").fetchall()])
         #We fetch all modifications here for temporary storage because it is VERY expensive to query peptide by peptide (3 seconds per 100 on my 500 MB test file, with 300,000 scans that's horrid)
         sql = 'select pam.PeptideID, GROUP_CONCAT(pam.AminoAcidModificationID), GROUP_CONCAT(pam.Position) from peptidesaminoacidmodifications pam GROUP BY pam.PeptideID'
@@ -1216,7 +1236,7 @@ class ThermoMSFIterator(templates.GenericIterator, GenericProteomicIterator):
 
     def getSILACLabels(self):
         labels = {}
-        import HTMLParser
+        from six.moves import html_parser as HTMLParser
         html_parser = HTMLParser.HTMLParser()
         if self.version == 1:
             sql = 'select ParameterValue from processingnodeparameters where ParameterName == "QuantificationMethod"'
@@ -1249,7 +1269,7 @@ class ThermoMSFIterator(templates.GenericIterator, GenericProteomicIterator):
         self.cur.execute(sql)
         for chroma in self.cur.fetchall():
             sInfo = chroma[2]
-            fp = StringIO.StringIO(sInfo)
+            fp = six.BytesIO(sInfo)
             zf = zipfile.ZipFile(fp, 'r')
             for j in zf.namelist():
                 msInfo = zf.read(j)
@@ -1331,15 +1351,13 @@ class ThermoMSFIterator(templates.GenericIterator, GenericProteomicIterator):
 
     def decompressScanInfo(self, scanObj, zip):
         sInfo = zip
-        fp = StringIO.StringIO(sInfo)
+        fp = six.BytesIO(sInfo)
         zf = zipfile.ZipFile(fp, 'r')
         for j in zf.namelist():
-            msInfo = zf.read(j)
+            msInfo = zf.read(j).decode()
             msStr = msInfo.split('\n')
-            #sIO = StringIO.StringIO('\n'.join(msStr[1:]))
             stage = 0
             #this is dirty, but unfortunately the fastest method at the moment
-            # print msInfo
             ms1_scans = []
             for row in msStr:
                 if stage == 0:
@@ -1412,12 +1430,13 @@ class ThermoMSFIterator(templates.GenericIterator, GenericProteomicIterator):
             mods = self.mods.get(int(pepId))
             if mods is not None:
                 for modId, modPosition in zip(mods[0].split(','),mods[1].split(',')):
-                    modEntry = self.modTable[int(modId)]
+                    modEntry = self.modTable[str(modId)]
                     scanObj.addModification(sequence[int(modPosition)], modPosition, modEntry[1], modEntry[0])
             tmods = self.tmods.get(int(pepId), [])
-            for modId in tmods:
-                modEntry = self.modTable[int(modId)]
-                scanObj.addModification('[', 0, modEntry[1], modEntry[0])
+            for modIds in tmods:
+                for modId in modIds.split(','):
+                    modEntry = self.modTable[str(modId)]
+                    scanObj.addModification('[', 0, modEntry[1], modEntry[0])
             scanObj.peptide = sequence
             scanObj.rank = searchRank
             scanObj.confidence = confidence
@@ -1452,13 +1471,14 @@ class ThermoMSFIterator(templates.GenericIterator, GenericProteomicIterator):
             mods = self.mods.get(int(pid))
             if mods is not None:
                 for modId, modPosition in zip(mods[0].split(','),mods[1].split(',')):
-                    modEntry = self.modTable[int(modId)]
+                    modEntry = self.modTable[str(modId)]
                     scanObj.addModification(peptide[int(modPosition)], modPosition, modEntry[1], modEntry[0])
             tmods = self.tmods.get(int(pid))
             if tmods is not None:
-                for modId in tmods:
-                    modEntry = self.modTable[int(modId)]
-                    scanObj.addModification('[', 0, modEntry[1], modEntry[0])
+                for modIds in tmods:
+                    for modId in modIds.split(','):
+                        modEntry = self.modTable[str(modId)]
+                        scanObj.addModification('[', 0, modEntry[1], modEntry[0])
         scanObj.peptide = peptide
         if self.decompressScanInfo(scanObj, i[0]):
             return scanObj
@@ -1530,7 +1550,7 @@ class MQIterator(templates.GenericIterator, GenericProteomicIterator):
         self.cur.execute(sql)
         for chroma in self.cur.fetchall():
             sInfo = chroma[2]
-            fp = StringIO.StringIO(sInfo)
+            fp = six.BytesIO(sInfo)
             zf = zipfile.ZipFile(fp, 'r')
             for j in zf.namelist():
                 msInfo = zf.read(j)
@@ -1598,15 +1618,13 @@ class MQIterator(templates.GenericIterator, GenericProteomicIterator):
 
     def decompressScanInfo(self, scanObj, zip):
         sInfo = zip
-        fp = StringIO.StringIO(sInfo)
+        fp = six.BytesIO(sInfo)
         zf = zipfile.ZipFile(fp, 'r')
         for j in zf.namelist():
             msInfo = zf.read(j)
             msStr = msInfo.split('\n')
-            #sIO = StringIO.StringIO('\n'.join(msStr[1:]))
             stage = 0
             #this is dirty, but unfortunately the fastest method at the moment
-            # print msInfo
             ms1_scans = []
             for row in msStr:
                 if stage == 0:
@@ -1651,7 +1669,6 @@ class MQIterator(templates.GenericIterator, GenericProteomicIterator):
                     elif '</PeakCentroids>' in row:
                         break
         if msInfo:
-	    # print scanObj.scans
             if master_scan != "-1" and ms1_scans:
                 master_scanobj = self.master_scans.get(master_scan, None)
                 if master_scanobj is None:
@@ -1680,9 +1697,8 @@ class MQIterator(templates.GenericIterator, GenericProteomicIterator):
             scanObj = PeptideObject()
             try:
                 mods = self.mods[int(pepId)]
-                # print mods
                 for modId, modPosition in zip(mods[0].split(','),mods[1].split(',')):
-                    modEntry = self.modTable[int(modId)]
+                    modEntry = self.modTable[str(modId)]
                     scanObj.addModification(sequence[int(modPosition)], modPosition, modEntry[1], modEntry[0])
             except KeyError:
                 pass
